//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// IBChess
// Copyright (C) 2002 by Ihor Bobak, ibobak@ukrpost.net
//------------------------------------------------------------------------------
// the description of the board (game position)
// !!!!!!!!!!!!!!!!!
// the squares of the board are numbered from 0 to 63 starting from the upper-left
// corner (square a8) to the lower-right (square h1)
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
unit board;

interface

uses global, move;

type
  TBoard = class
  private
    // material factor of the estimation function 
    MaterialValue: array[0..1] of integer;
    // current player
    FCurrentPlayer: integer;
    // bitboards for each piece type
    FBitBoards: array[0..ALL_BITBOARDS-1] of TBitBoard;
    // was the castring already done?
    FHasCastled: array[0..2] of boolean;
    // is there a possibility of castling
    FCastlingPossibility: array[0..3] of boolean;
    // en-passant pawn bitboard
    FEnPassantPawn: TBitBoard;
    // extra-kings: used for calculation of doing the castling 
    FExtraKings: array[0..1] of TBitBoard;
    FTotalPiecesByColor: array[0..1] of integer;
    // put the figure on the board
    function AddPiece(whichSquare, whichPiece: integer): boolean;
    // deletes the figure from the board
    function RemovePiece(whichSquare: integer; whichPiece: integer): boolean;
    // clears the board
    function EmptyBoard: boolean;
    // set the bit-board for en-passant pawn
    function SetEnPassantPawnSquare(square: integer): boolean;
    // property functions
    function GetBitBoard(Index: integer): TBitBoard;
    function GetHasCastled(Index: integer): boolean;
    function GetCastlingPossibility(Index: integer): boolean;
    function GetExtraKings(Index: integer): TBitBoard;
    function GetTotalPiecesByColor(Index: integer): integer;
    procedure SetExtraKings(side: integer; val: TBitBoard);
  public
    constructor Create;

    property BitBoards[Index: integer]: TBitBoard read GetBitBoard;
    property CastlingPossibility[Index: integer]: boolean read GetCastlingPossibility;
    property CurrentPlayer: integer read FCurrentPlayer;
    property EnPassantPawn: TBitBoard read FEnPassantPawn;
    property ExtraKings[Index: integer]: TBitBoard read GetExtraKings;
    property HasCastled[Index: integer]: boolean read GetHasCastled;
    property TotalPiecesByColor[Index: integer]: integer read GetTotalPiecesByColor;

    // do the move
    procedure ApplyMove(theMove: TMove);
    // clear the extra-kings
    procedure ClearExtraKings(side: integer); //must be public
    // clone the board squares 
    procedure Clone(target: TBoard);
    // set the initial state of the board
    procedure InitializeBoard;
    // load the board from the file
    procedure Load(fileName: string);
    // print the board
    procedure Print;
    // switch the player
    procedure SwitchSides;
    // estimate material overweight
    function EvalMaterialDiff(side: integer): integer;
    // find a black piece on the given square
    function FindBlackPiece(square: integer): integer;
    // find a white piece on the given square
    function FindWhitePiece(square: integer): integer;
    // hash-key and hash-lock generation
    function HashLock: integer;
    function HashKey: integer;
    function toString: string;
  end;

implementation

var
  // random values of components for counting the hash-function
  HashKeyBoardComponents, HashLockBoardComponents: array[0..ALL_PIECES-1, 0.. ALL_SQUARES-1] of integer;
  HashKeyEnpassantComponents, HashLockEnpassantComponents: array [0..ALL_SQUARES-1] of integer;
  HashKeyPlayerComponents, HashLockPlayerComponents: array[SIDE_WHITE..SIDE_BLACK] of integer; 
  HashKeyCastlingComponents, HashLockCastlingComponents: array[0..3] of integer;

constructor TBoard.Create;
begin
  InitializeBoard;
end;

// extra kings are extra 1's in the king's bitboard in 
// order to control the possibility of castling
procedure TBoard.SetExtraKings(side: integer; val: TBitBoard);
begin
  FExtraKings[side] := val;
  FBitBoards[KING + side] := FBitBoards[KING + side]  or FExtraKings[side];
  FBitBoards[ALL_PIECES + side] := FBitBoards[ALL_PIECES + side] or FExtraKings[side];
end;

procedure TBoard.ClearExtraKings(side: integer);
begin
  FBitBoards[KING + side] := FBitBoards[KING + side] xor FExtraKings[side];
  FBitBoards[ALL_PIECES + side] := FBitBoards[ALL_PIECES + side] xor FExtraKings[side];
  FBitBoards[ALL_PIECES + side] := FBitBoards[ALL_PIECES + side] or FBitBoards[ROOK + side];
  FExtraKings[side] := 0;
end;

// what black piece is on the square?
function TBoard.FindBlackPiece(square: integer): integer;
begin
  if (FBitBoards[BLACK_KING] and SquareBits[square]) <> 0 then
    Result := BLACK_KING
  else if (FBitBoards[BLACK_QUEEN] and SquareBits[square]) <> 0 then
    Result := BLACK_QUEEN
  else if (FBitBoards[BLACK_ROOK] and SquareBits[square]) <> 0 then
    Result := BLACK_ROOK
  else if (FBitBoards[BLACK_KNIGHT] and SquareBits[square]) <> 0 then
    Result := BLACK_KNIGHT
  else if (FBitBoards[BLACK_BISHOP] and SquareBits[square]) <> 0 then
    Result := BLACK_BISHOP
  else if (FBitBoards[BLACK_PAWN] and SquareBits[square]) <> 0 then
    Result := BLACK_PAWN
  else
    Result := EMPTY_SQUARE;
end;

// what white piece is on the square?
function TBoard.FindWhitePiece(square: integer): integer;
begin
  if (FBitBoards[WHITE_KING] and SquareBits[square]) <> 0 then
    Result := WHITE_KING
  else if (FBitBoards[WHITE_QUEEN] and SquareBits[square]) <> 0 then
    Result := WHITE_QUEEN
  else if (FBitBoards[WHITE_ROOK] and SquareBits[square]) <> 0 then
    Result := WHITE_ROOK
  else if FBitBoards[WHITE_KNIGHT] and SquareBits[square] <> 0 then
    Result := WHITE_KNIGHT
  else if (FBitBoards[WHITE_BISHOP] and SquareBits[square]) <> 0 then
    Result := WHITE_BISHOP
  else if (FBitBoards[WHITE_PAWN] and SquareBits[square]) <> 0 then
    Result := WHITE_PAWN
  else
    Result := EMPTY_SQUARE;
end;

// makes the copy of the board
procedure TBoard.Clone(target: TBoard);
var
  i: integer;
begin
  FEnPassantPawn := target.EnPassantPawn;
  for i := 0 to 3 do
    FCastlingPossibility[i] := target.CastlingPossibility[i];
  for  i := 0 to ALL_BITBOARDS-1 do
    FBitBoards[i] := target.BitBoards[i];
  MaterialValue[0] := target.MaterialValue[0];
  MaterialValue[1] := target.MaterialValue[1];
  FExtraKings[0] := target.ExtraKings[0];
  FExtraKings[1] := target.ExtraKings[1];
  FHasCastled[0] := target.HasCastled[0];
  FHasCastled[1] := target.HasCastled[1];
  FCurrentPlayer := target.CurrentPlayer;
end;

// gives the right of making move to the opponent
procedure TBoard.SwitchSides;
begin
  FCurrentPlayer := (FCurrentPlayer + 1) mod 2;
end;

// finding the key by Zobrist algorithm
// the whole board is described be:
// 1) the bitboards of the pieces
// 2) the bitboard of the en-passant pawm
// 3) the player, whose move is now
// 4) castling possibility
// Two boards are different in the case when at least one
// of these conditions is not true.
function TBoard.HashKey: integer;
var
  hash: integer;
  tmp: TBitBoard;
  currPiece, currSquare, i: integer;
begin
  hash := 0;
  // pieces on the board, including en-passant pawn
  for currPiece := 0 to ALL_PIECES-1 do
  begin
    tmp := BitBoards[currPiece];
    for currSquare := 0 to ALL_SQUARES-1 do
      if (tmp and SquareBits[currSquare]) <> 0 then
        hash := hash xor HashKeyBoardComponents[currPiece][currSquare];
  end;

  //take into account the en-passant pawn
  if FEnPassantPawn <> 0 then
    for currSquare := 0 to ALL_SQUARES-1 do
      if (FEnPassantPawn and SquareBits[currSquare]) <> 0 then
        hash := hash xor HashKeyEnpassantComponents[currSquare];

  // the move of the player
  hash := hash xor HashKeyPlayerComponents[FCurrentPlayer];
  // castling possibility
  for i:=0 to 3 do
    if FCastlingPossibility[i] then
      hash := hash xor HashKeyCastlingComponents[i];
  Result := hash;
end;


// counts the hash-lock by the same algorithm as hash-key
function TBoard.HashLock: integer;
var
  hash: integer;
  tmp: TBitBoard;
  currPiece, currSquare, i: integer;
begin
  hash := 0;
  for currPiece := 0 to ALL_PIECES-1 do
  begin
    tmp := BitBoards[currPiece];
    for currSquare := 0 to ALL_SQUARES-1 do
      if (tmp and SquareBits[currSquare]) <> 0 then
        hash := hash xor HashLockBoardComponents[currPiece][currSquare];
  end;
  if FEnPassantPawn <> 0 then
    for currSquare := 0 to ALL_SQUARES-1 do
      if (FEnPassantPawn and SquareBits[currSquare]) <> 0 then
        hash := hash xor HashLockEnpassantComponents[currSquare];
  hash := hash xor HashLockPlayerComponents[CurrentPlayer];
  for i:=0 to 3 do
    if FCastlingPossibility[i] then
      hash := hash xor HashLockCastlingComponents[i];
  Result := hash;
end;


// applies the move to the board
procedure TBoard.ApplyMove(theMove: TMove);
var
  isPromotion: boolean;
  promotionType: integer;
  moveWithoutPromotion: integer;
  side: integer;
  theRook: integer;
  color: integer;
begin
  isPromotion := theMove.MoveType >= MOVE_PROMOTION_KNIGHT;
  moveWithoutPromotion := theMove.MoveType and NO_PROMOTION_MASK ;
  side := theMove.MovingPiece mod 2;

  case moveWithoutPromotion of
    MOVE_NORMAL:
    begin
      RemovePiece(theMove.SourceSquare, theMove.MovingPiece);
      AddPiece(theMove.DestinationSquare, theMove.MovingPiece);
    end;

    MOVE_CAPTURE_ORDINARY:
    begin
      RemovePiece(theMove.SourceSquare, theMove.MovingPiece);
      RemovePiece(theMove.DestinationSquare, theMove.CapturedPiece);
      AddPiece(theMove.DestinationSquare, theMove.MovingPiece);
      case theMove.DestinationSquare of
        0:  FCastlingPossibility[CASTLE_QUEENSIDE + SIDE_BLACK] := false;
        7:  FCastlingPossibility[CASTLE_KINGSIDE + SIDE_BLACK]  := false;
        56: FCastlingPossibility[CASTLE_QUEENSIDE + SIDE_WHITE] := false;
        63: FCastlingPossibility[CASTLE_KINGSIDE + SIDE_WHITE]  := false;
      end;
    end;

    MOVE_CAPTURE_EN_PASSANT:
    begin
      RemovePiece(theMove.SourceSquare, theMove.MovingPiece);
      AddPiece(theMove.DestinationSquare, theMove.MovingPiece);
      if side = SIDE_WHITE then
        RemovePiece(theMove.DestinationSquare + 8, theMove.CapturedPiece)
      else
        RemovePiece(theMove.DestinationSquare - 8, theMove.CapturedPiece);
    end;

    MOVE_CASTLING_QUEENSIDE:
    begin
      RemovePiece(theMove.SourceSquare, theMove.MovingPiece);
      AddPiece(theMove.DestinationSquare, theMove.MovingPiece);
      theRook := ROOK + side;
      RemovePiece(theMove.SourceSquare - 4, theRook);
      AddPiece(theMove.SourceSquare - 1, theRook);
      if side = SIDE_WHITE then
        SetExtraKings(side, EXTRAKINGS_WHITE_QUEENSIDE)
      else
        SetExtraKings(side, EXTRAKINGS_BLACK_QUEENSIDE);
      FHasCastled[side] := true;
    end;

    MOVE_CASTLING_KINGSIDE:
    begin
      RemovePiece(theMove.SourceSquare, theMove.MovingPiece);
      AddPiece(theMove.DestinationSquare, theMove.MovingPiece);
      theRook := ROOK + side;
      RemovePiece(theMove.SourceSquare + 3, theRook);
      AddPiece(theMove.SourceSquare + 1, theRook);
      if side = SIDE_WHITE then
        SetExtraKings(side, EXTRAKINGS_WHITE_KINGSIDE)
      else
        SetExtraKings(side, EXTRAKINGS_BLACK_KINGSIDE);
      FHasCastled[side] := true;
    end;

    MOVE_RESIGN:
    begin
      // to be done later
    end;

    MOVE_STALEMATE:
    begin
      writeln('Stalemate - Game is a draw.');
    end;
  end;

  if isPromotion then
  begin
    promotionType := (theMove.MoveType and PROMOTION_MASK);
    color := (theMove.MovingPiece mod 2);
    case promotionType of
      MOVE_PROMOTION_KNIGHT:
      begin
        RemovePiece(theMove.DestinationSquare, theMove.MovingPiece);
        AddPiece(theMove.DestinationSquare, KNIGHT + color);
      end;
      MOVE_PROMOTION_BISHOP:
      begin
        RemovePiece(theMove.DestinationSquare, theMove.MovingPiece);
        AddPiece(theMove.DestinationSquare, BISHOP + color);
      end;
      MOVE_PROMOTION_ROOK:
      begin
        RemovePiece(theMove.DestinationSquare, theMove.MovingPiece);
        AddPiece(theMove.DestinationSquare, ROOK + color);
      end;
      MOVE_PROMOTION_QUEEN:
      begin
        RemovePiece(theMove.DestinationSquare, theMove.MovingPiece);
        AddPiece(theMove.DestinationSquare, QUEEN + color);
      end;
    end;
  end;

  if (theMove.MovingPiece = WHITE_PAWN) and
       (theMove.SourceSquare - theMove.DestinationSquare = 16) then
    SetEnPassantPawnSquare(theMove.DestinationSquare + 8)
  else if (theMove.MovingPiece = BLACK_PAWN) and
            (theMove.DestinationSquare - theMove.SourceSquare = 16) then
    SetEnPassantPawnSquare(theMove.SourceSquare + 8)
  else
    FEnPassantPawn := 0;

  case theMove.MovingPiece of
    WHITE_KING:
    begin
      FCastlingPossibility[CASTLE_KINGSIDE + SIDE_WHITE] := false;
      FCastlingPossibility[CASTLE_QUEENSIDE + SIDE_WHITE] := false;
    end;
    BLACK_KING:
    begin
      FCastlingPossibility[CASTLE_KINGSIDE + SIDE_BLACK] := false;
      FCastlingPossibility[CASTLE_QUEENSIDE + SIDE_BLACK] := false;
    end;
  end;

  case theMove.SourceSquare of
    0:  FCastlingPossibility[CASTLE_QUEENSIDE + SIDE_BLACK] := false;
    7:  FCastlingPossibility[CASTLE_KINGSIDE + SIDE_BLACK]  := false;
    56: FCastlingPossibility[CASTLE_QUEENSIDE + SIDE_WHITE] := false;
    63: FCastlingPossibility[CASTLE_KINGSIDE + SIDE_WHITE]  := false;
  end;

  SwitchSides;
end;


// reads the board from the file. This function was used
// with testing purposes 
procedure TBoard.Load(fileName: string);
var
  f: textfile;
  rank, file1: integer;
  line: string;
  piece: integer;
  function StrToPiece(str: string): integer;
  var
    color, f: integer;
  begin
    if str[1] = '-' then
    begin
      Result := EMPTY_SQUARE;
      exit;
    end;
    if str[1] = 'B' then
      color := SIDE_BLACK
    else color := SIDE_WHITE;
    case str[2] of
      'R': f := ROOK;
      'N': f := KNIGHT;
      'Q': f := QUEEN;
      'K': f := KING;
      'B': f := BISHOP;
      'P': f := PAWN;
    end;
    Result := f + color;
  end;
begin
  assignFile(f, fileName);
  reset(f);
  EmptyBoard;
  for rank:=0 to 7 do
  begin
    readln(f, line);
    for file1 := 0 to 7 do
    begin
      piece := StrToPiece(copy(line, file1*3+1, 2));
      if piece <> EMPTY_SQUARE then
        AddPiece(rank shl 3 + file1, piece);
    end;
    if rank <> 7 then
      readln(f); //read empty line
  end;
  closeFile(f);
end;

// gives the material estimate of the player side:
// used in the estimation function
function TBoard.EvalMaterialDiff(side: integer): integer;
var
  otherSide: integer;
begin
  otherSide := (side + 1) mod 2;
  Result := MaterialValue[side] - MaterialValue[otherSide];
end;

// initializes the board
procedure TBoard.InitializeBoard;
var
  i: integer;
begin
  EmptyBoard;
  AddPiece(0, BLACK_ROOK);
  AddPiece(1, BLACK_KNIGHT);
  AddPiece(2, BLACK_BISHOP);
  AddPiece(3, BLACK_QUEEN);
  AddPiece(4, BLACK_KING);
  AddPiece(5, BLACK_BISHOP);
  AddPiece(6, BLACK_KNIGHT);
  AddPiece(7, BLACK_ROOK);
  for i := 8 to 15 do
    AddPiece(i, BLACK_PAWN);
  for i := 48 to 55 do
    AddPiece(i, WHITE_PAWN);
  AddPiece(56, WHITE_ROOK);
  AddPiece(57, WHITE_KNIGHT);
  AddPiece(58, WHITE_BISHOP);
  AddPiece(59, WHITE_QUEEN);
  AddPiece(60, WHITE_KING);
  AddPiece(61, WHITE_BISHOP);
  AddPiece(62, WHITE_KNIGHT);
  AddPiece(63, WHITE_ROOK);
  for i := 0 to 3 do
    FCastlingPossibility[i] := true;
  FHasCastled[0] := false;
  FHasCastled[1] := false;
  FEnPassantPawn := 0;
  FCurrentPlayer := SIDE_WHITE;
  FTotalPiecesByColor[0] := 16;
  FTotalPiecesByColor[1] := 16;
end;

// puts the piece on the board to the specified square
function TBoard.AddPiece(whichSquare, whichPiece: integer): boolean;
begin
  FBitBoards[whichPiece] := FBitBoards[whichPiece] or SquareBits[whichSquare];
  FBitBoards[ALL_PIECES + (whichPiece mod 2)] :=
  FBitBoards[ALL_PIECES + (whichPiece mod 2)] or SquareBits[whichSquare];
  inc(MaterialValue[whichPiece mod 2], PieceValues[whichPiece]);
  inc(FTotalPiecesByColor[whichPiece mod 2]);
  Result := true;
end;

// takes away the figure from the board 
function TBoard.RemovePiece(whichSquare: integer; whichPiece: integer): boolean;
begin
  FBitBoards[whichPiece] := FBitBoards[whichPiece] xor SquareBits[whichSquare];
  FBitBoards[ALL_PIECES + (whichPiece mod 2)] :=
  FBitBoards[ALL_PIECES + (whichPiece mod 2)] xor SquareBits[whichSquare];
  dec(MaterialValue[whichPiece mod 2], PieceValues[whichPiece]);
  dec(FTotalPiecesByColor[whichPiece mod 2]);
  Result := true;
end;


function TBoard.GetBitBoard(Index: integer): TBitBoard;
begin
  Result := FBitBoards[Index];
end;

function TBoard.GetHasCastled(Index: integer): boolean;
begin
  Result := FHasCastled[Index];
end;

function TBoard.GetCastlingPossibility(Index: integer): boolean;
begin
  Result := FCastlingPossibility[Index];
end;

function TBoard.GetExtraKings(Index: integer): TBitBoard;
begin
  Result := FExtraKings[Index];
end;

function TBoard.GetTotalPiecesByColor(Index: integer): integer;
begin
  Result := FTotalPiecesByColor[Index];
end;

// clears the board
function TBoard.EmptyBoard: boolean;
var
  i: integer;
begin
  for i := 0 to ALL_BITBOARDS-1 do
    FBitBoards[i] := 0;
  FExtraKings[0] := 0;
  FExtraKings[1] := 0;
  FEnPassantPawn := 0;
  MaterialValue[0] := 0;
  MaterialValue[1] := 0;
  Result := true;
  FTotalPiecesByColor[0] := 0;
  FTotalPiecesByColor[1] := 0;
end;

// sets the bitboard for the en-passant pawn
function TBoard.SetEnPassantPawnSquare(square: integer): boolean;
begin
  FEnPassantPawn := 0;
  FEnPassantPawn := FEnPassantPawn or SquareBits[square];
  Result := true;
end;


procedure TBoard.Print;
var
  line, col: integer;
  bits: TBitBoard;
  piece: integer;
begin
  for line := 0 to 7 do
  begin
    writeln(' ---- ---- ---- ---- ---- ---- ---- ---- ');
    writeln('|    |    |    |    |    |    |    |    |');
    for col := 0 to 7 do
    begin
      bits := SquareBits[line * 8 + col];
      piece := 0;
      while (piece < ALL_PIECES) and ((bits and BitBoards[piece]) = 0) do
        inc(piece);
      if (piece = BLACK_KING) and ((ExtraKings[SIDE_BLACK] and SquareBits[line * 8 + col]) <> 0) then
        piece := EMPTY_SQUARE;
      if (piece = WHITE_KING) and ((ExtraKings[SIDE_WHITE] and SquareBits[line * 8 + col]) <> 0) then
        piece := EMPTY_SQUARE;
      write('| ' + PieceStrings[piece] + ' ');
    end;
    writeln('|');
    writeln('|    |    |    |    |    |    |    |    |');
  end;
    writeln(' ---- ---- ---- ---- ---- ---- ---- ---- ');
end;

function TBoard.toString: string;
var
  line, col: integer;
  bits: TBitBoard;
  piece: integer;
begin
  Result := '';
  for line := 0 to 7 do
  begin
    Result :=  Result + '-----------------------------------------'#13#10;
    Result :=  Result + '|    |    |    |    |    |    |    |    |'#13#10;
    for col := 0 to 7 do
    begin
      bits := SquareBits[line * 8 + col];
      piece := 0;
      while (piece < ALL_PIECES) and ((bits and BitBoards[piece]) = 0) do
        inc(piece);
      if (piece = BLACK_KING) and ((ExtraKings[SIDE_BLACK] and SquareBits[line * 8 + col]) <> 0) then
        piece := EMPTY_SQUARE;
      if (piece = WHITE_KING) and ((ExtraKings[SIDE_WHITE] and SquareBits[line * 8 + col]) <> 0) then
        piece := EMPTY_SQUARE;

      // Show the piece
      Result := Result + '| ' + PieceStrings[piece] + ' ';
    end;
    Result := Result +  '|'#13#10;
    Result := Result +  '|    |    |    |    |    |    |    |    |'#13#10;
  end;
  Result := Result +'-----------------------------------------'#13#10;
end;        

var
  i, j: integer;

initialization

  // initialization of hash components 
  randomize;
  for i := 0 to ALL_PIECES-1 do
    for j := 0 to ALL_SQUARES-1 do
    begin
      HashKeyBoardComponents[i][j] := random(Maxint);
      HashLockBoardComponents[i][j] := random(Maxint);
    end;

  for j := 0 to ALL_SQUARES-1 do
    HashKeyEnpassantComponents[j] := random(MaxInt);

  HashKeyPlayerComponents[SIDE_WHITE] := random(Maxint);
  HashKeyPlayerComponents[SIDE_BLACK] := random(Maxint);
  HashLockPlayerComponents[SIDE_WHITE] := random(Maxint);
  HashLockPlayerComponents[SIDE_BLACK] := random(Maxint);

  for i:=0 to 3 do
  begin
    HashKeyCastlingComponents[i] := random(MaxInt);
    HashLockCastlingComponents[i] := random(MaxInt);
  end;

  // the values of figures on the board
  PieceValues[WHITE_PAWN] := 100;
  PieceValues[BLACK_PAWN] := 100;
  PieceValues[WHITE_KNIGHT] := 300;
  PieceValues[BLACK_KNIGHT] := 300;
  PieceValues[WHITE_BISHOP] := 350;
  PieceValues[BLACK_BISHOP] := 350;
  PieceValues[WHITE_ROOK] := 500;
  PieceValues[BLACK_ROOK] := 500;
  PieceValues[BLACK_QUEEN] := 900;
  PieceValues[WHITE_QUEEN] := 900;
  PieceValues[WHITE_KING] := 2000;
  PieceValues[BLACK_KING] := 2000;
  PieceValues[EMPTY_SQUARE] := 0;
end.
