//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// IBChess
// Copyright
// (C) 2002 by Ihor Bobak, ibobak@ukrpost.net
//------------------------------------------------------------------------------
// the brain of the program
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
unit SearchAgent;

interface

uses SysUtils, global, tables, move, movegenerator, board, OpeningBook, boardeval;

type
  // this class is the branch responsible for the search
  TAISearchAgent = class
  protected
    TransTable: TTranspositionTable; // 
    HistoryTable: THistoryTable; // historical heuristics
    Evaluator: TBoardEvaluator; // the position estimato
    FromWhosePerspective: integer; // the player, from whose perspective the evaluation is done
    OpeningBook: TOpeningBook; // the opening book

    // the brain of the program is here
    function AlphaBeta(nodeType: boolean; theBoard: TBoard; depth, alpha, beta: integer;
      BestMoveToFind: TMove; recursionDepth: integer): integer;

    // this search works the same as alpha-beta, but consideres only captures
    function QuiescenceSearch(nodeType: boolean; theBoard: TBoard; alpha, beta: integer;
      recursionDepth: integer): integer;
    // move ordering
    procedure SortMoves(movegen: TMoveGenerator; bestMoveFountInTT: TMove; CurrentPlayer: integer; quiescence: boolean = false);
    // MTDF-search; calls AlphaBeta
    function MTDF( theBoard: TBoard; target, depth: integer; var BestMove: TMove): integer;
  public
    constructor Create(AOpeningBook: TOpeningBook);
    destructor Destroy; override;
    // Iterative Deepness MTD-f
    function PickBestMove(theBoard: TBoard): TMove;
  end;

implementation

constructor TAISearchAgent.Create(AOpeningBook: TOpeningBook);
begin
  OpeningBook := AOpeningBook;
  TransTable := TTranspositionTable.Create;
  HistoryTable := THistoryTable.Create;
  Evaluator := TBoardEvaluator.Create;
end;


destructor TAISearchAgent.Destroy;
begin
  Evaluator.Free;
  HistoryTable.Free;
  TransTable.Free;
end;


// alpha-beta search: called from inside the MTDF
function TAISearchAgent.AlphaBeta(nodeType: boolean; theBoard: TBoard;
  depth, alpha, beta: integer; BestMoveToFind: TMove; recursionDepth: integer): integer;
var
  lowerbound, upperbound: integer;
  tableMove: TMove;
  tableDepth: integer;
  g, a, b: integer;
  movegen: TMoveGenerator;
  currentMove, bestCurrentMove: TMove;
  boardAfterMove: TBoard;
  i, eval: integer;
label ex, fin;
begin
  // variable initialization
  bestCurrentMove := nil;
  lowerbound := ALPHABETA_MINVAL;
  upperbound := ALPHABETA_MAXVAL;
  tableMove := TMove.Create;
  movegen := TMoveGenerator.Create;
  boardAfterMove := TBoard.Create;

  //checking the position for legality
  if movegen.CanWeBeatHisKing(theBoard) then
  begin
    g := ALPHABETA_ILLEGAL;
    goto ex;
  end;

  // trying to find the position in TT
  if TransTable.LookupBoard(theBoard, lowerbound, upperbound, tableMove, tableDepth) then
  begin
    // if this position was investigated to enough deepness,
    if tableDepth >= depth then
    begin
      // check whether it is possible to cut off the search
      if lowerbound >= beta then
      begin
        g := lowerbound;
        if BestMoveToFind<>nil then BestMoveToFind.Clone(tableMove);
        goto ex;
      end;
      if upperbound <= alpha then
      begin
        g := upperbound;
        if BestMoveToFind<>nil then BestMoveToFind.Clone(tableMove);
        goto ex;
      end;
      // correction of the bounds
      alpha := max(alpha, lowerbound);
      beta  := min(beta, upperbound);
    end else begin
      //we have lookup move
    end;
  end
  else
    FreeAndNil(tableMove);

  // if we approached the terminal position
  if depth = 0 then
  begin
    // then instead of the estimation of the situation return the result of QuiescenceSearch
    g := QuiescenceSearch(nodeType, theBoard, alpha, beta, recursionDepth); //we do not add 1 here !
    goto ex;
  end;

  // generate all possible moves and order them
  movegen.GenerateLegalMoves(theBoard);
  SortMoves(movegen, tableMove, theBoard.CurrentPlayer);

  // the main part of alpha-beta
  case nodeType of
    MAXNODE:
    BEGIN
      g := ALPHABETA_MINVAL; // the current value of the node
      a := alpha;
      // iterate through all nodes
      for i := 0 to movegen.Moves.Count-1 do
      begin
        currentMove := TMove(movegen.Moves[i]); // current move
        boardAfterMove.Clone(theBoard); //  copy of the board
        boardAfterMove.ApplyMove(currentMove);
        // recursively call alpha-beta
        eval := AlphaBeta(not nodeType, boardAfterMove, depth-1, a, beta, nil, recursionDepth+1);
        if eval = ALPHABETA_ILLEGAL then
          continue;

        // if the found value is better than the current
        if eval > g then
        begin
          g := eval; // then store it in g
          bestCurrentMove := currentMove; // and remember the last move
        end;

        // if g >= beta, perform the cut-off
        if g >= beta then
          goto fin;
        a := max(a, g);
      end;

      // if there were no move, then the position is either checkmate,
      // or stalemate
      if g <= ALPHABETA_MINVAL then
      begin
        // if the king is under a check, then this is checkmate
        if movegen.IsInCheck(theBoard) then
        begin
          g := g + depth; // the result is - ALPHABETA_MINVAL + deepness
          if BestMoveToFind<>nil then BestMoveToFind.MoveType := MOVE_RESIGN;
        end else begin
          g := 0; // the result is 0
          if BestMoveToFind<>nil then BestMoveToFind.MoveType := MOVE_STALEMATE
        end;
        goto ex;
      end;
    END;

    MINNODE:
    BEGIN
      g := ALPHABETA_MAXVAL;
      b := beta;
      for i := 0 to movegen.Moves.Count-1 do
      begin
        currentMove := TMove(movegen.Moves[i]);
        boardAfterMove.Clone(theBoard);
        boardAfterMove.ApplyMove(currentMove);
        eval := AlphaBeta(not nodeType, boardAfterMove, depth-1, alpha, b, nil, recursionDepth+1);
        if eval = ALPHABETA_ILLEGAL then
          continue;

        if eval < g then
        begin
          g := eval;
          bestCurrentMove := currentMove;
        end;

        if g <= alpha then
          goto fin;
        b := min(b, g);
      end;

      if g >= ALPHABETA_MAXVAL then
      begin
        if movegen.IsInCheck(theBoard) then
        begin
          g := g - depth;
          if BestMoveToFind<>nil then BestMoveToFind.MoveType := MOVE_RESIGN;
        end else begin
          g := 0;
          if BestMoveToFind<>nil then BestMoveToFind.MoveType := MOVE_STALEMATE
        end;
        goto ex;
      end;
    END;
  end;

fin:
  // find the bounds of this node
  if g <= alpha then upperbound := g;
  if (g >= alpha) and (g <= beta) then
  begin
    lowerbound := g;
    upperbound := g;
  end;
  if g >= beta then lowerbound := g;
  // and store them in TT
  TransTable.StoreBoard(theBoard, lowerbound, upperbound, depth, bestCurrentMove);
  if BestMoveToFind<>nil then BestMoveToFind.Clone(bestCurrentMove);
  // index the values of the historical heuristics table
  HistoryTable.AddCount(theBoard.CurrentPlayer, bestCurrentMove, depth);

ex:
  Result := g;
  // destroy all temporary objects
  if tableMove <> nil then
    tableMove.Free;
  movegen.Free;
  if boardAfterMove <> nil then
    boardAfterMove.Free;
end;


// almost the same search as alpha-beta, but among the captures
function TAISearchAgent.QuiescenceSearch(nodeType: boolean;
  theBoard: TBoard; alpha, beta, recursionDepth: integer): integer;
var
  lowerbound, upperbound: integer;
  tableMove: TMove;
  tableDepth: integer;
  g, a, b: integer;
  movegen: TMoveGenerator;
  currentMove, bestCurrentMove: TMove;
  boardAfterMove: TBoard;
  i, eval: integer;
  NilMove: TMove;
label ex, fin;
begin
  // local variable initialization
  NilMove := nil;
  bestCurrentMove := nil;
  lowerbound := ALPHABETA_MINVAL;
  upperbound := ALPHABETA_MAXVAL;
  tableMove := TMove.Create;
  movegen := TMoveGenerator.Create;
  boardAfterMove := TBoard.Create;

  //checking the position for legality
  if movegen.CanWeBeatHisKing(theBoard) then
  begin
    g := ALPHABETA_ILLEGAL;
    goto ex;
  end;

  // if we are under check, transfer control to alpha-beat
  if movegen.IsInCheck(theBoard) then
  begin
    g := AlphaBeta(nodeType, theBoard, integer(1), alpha, beta, nil, recursionDepth + 1);
    goto ex;
  end;

  // find the move in the TT
  if TransTable.LookupBoard(theBoard, lowerbound, upperbound, tableMove, tableDepth) then
  begin
    // if f- > beta - cut-off the search
    if lowerbound >= beta then
    begin
      g := lowerbound;
      goto ex;
    end;
    // if  f+ < alpha - the same
    if upperbound <= alpha then
    begin
      g := upperbound;
      goto ex;
    end;
    // bounds correction
    alpha := max(alpha, lowerbound);
    beta  := min(beta, upperbound);
  end
  else
    FreeAndNil(tableMove);

  // try to estimate thsi situation
  g := Evaluator.Evaluate(theBoard, fromWhosePerspective);
  if NodeType = MAXNODE then
  begin
    if g >= beta then
      goto ex;
    alpha := max(g, alpha);
  end
  else
  begin
    if g <= alpha then
      goto ex;
    beta := min(g, beta);
  end;

  // if there are no more captures, that this is a quiet position
  // and return the result
  if not movegen.GenerateCaptureMoves(theBoard) then
    goto fin;

  // sort the moves
  SortMoves(movegen, tableMove, theBoard.CurrentPlayer, true);

  case nodeType of
    MAXNODE:
    BEGIN
      // g starts moving from the position estimation, but not from  ALPHABETA_MINVAL
      // the rest code is very similar to what we have in alpha-beta
      a := alpha;
      for i := 0 to movegen.Moves.Count-1 do
      begin
        currentMove := TMove(movegen.Moves[i]);
        boardAfterMove.Clone(theBoard);
        boardAfterMove.ApplyMove(currentMove);

        eval := QuiescenceSearch(not nodeType, boardAfterMove, a, beta, recursionDepth + 1);
        if eval = ALPHABETA_ILLEGAL then
          continue;

        if eval > g then
        begin
          g := eval;
          bestCurrentMove := currentMove;
        end;

        if g >= beta then
          goto fin;
        a := max(a, g);
      end;

      if g <= ALPHABETA_MINVAL then
      begin
        if movegen.IsInCheck(theBoard) then
          g := g
        else
          g := 0;
        goto ex;
      end;
    END;

    MINNODE: //the same code for min-node
    BEGIN
      b := beta;
      for i := 0 to movegen.Moves.Count-1 do
      begin
        currentMove := TMove(movegen.Moves[i]);
        boardAfterMove.Clone(theBoard);
        boardAfterMove.ApplyMove(currentMove);

        eval := QuiescenceSearch(not nodeType, boardAfterMove, alpha, b, recursionDepth + 1);
        if eval = ALPHABETA_ILLEGAL then
          continue;

        if eval < g then
        begin
          g := eval;
          bestCurrentMove := currentMove;
        end;

        if g <= alpha then
          goto fin;
        b := min(b, g);
      end;

      if g >= ALPHABETA_MAXVAL then
      begin
        if movegen.IsInCheck(theBoard) then
          g := g
        else
          g := 0;
        goto ex;
      end;
    END;
  end;

fin:
  // find the bounds for this node
  if g <= alpha then upperbound := g;
  if (g >= alpha) and (g <= beta) then
  begin
    lowerbound := g;
    upperbound := g;
  end;
  if g >= beta then lowerbound := g;

  // if the search did not continue recursively
  if bestCurrentMove = nil then
  begin
    NilMove := TMove.Create; // then we create a null move
    TransTable.StoreBoard(theBoard, lowerbound, upperbound, 0, NilMove);
    NilMove.Free;
  end
  else
    // keep the move in 
    TransTable.StoreBoard(theBoard, lowerbound, upperbound, 0, bestCurrentMove);

ex:
  Result := g;

  if tableMove <> nil then
    tableMove.Free;
  movegen.Free;
  boardAfterMove.Free;
end;



// order of the moves:
// 1) at the  beginning must be the move from TT
// 2) then all the moves by the principle MVV/LVA
// 3) pieceful moves ordered by the historical heuristics
procedure TAISearchAgent.SortMoves(movegen: TMoveGenerator;
  bestMoveFountInTT: TMove;
  CurrentPlayer: integer;
  quiescence: boolean = false);

  // a part of Quick-Sort algorith for sorting the captures
  // this algorithm sorts in the non-incremental order.
  function PartitionCaptures(p, r: integer): integer;
  var
    i, j: integer;
    x: TMove;
    xvalue: integer;
  begin
    i := p - 1;
    j := r + 1;
    x := TMove(movegen.Moves[p]);
    // Most Valuable Victim/Least Valuable Attacker
    // the most value of king is  (2000).
    // its enough to soft by CapturedPiece * 2^12 - we'll get the
    // decreasing of Victim,  and by adding
    // (- PieceValues[x.MovingPiece]), we'll get the increase by Attacker;
    xvalue := PieceValues[x.CapturedPiece] shl 12 - PieceValues[x.MovingPiece];
    while true do
    begin
      repeat dec(j) until PieceValues[TMove(movegen.Moves[j]).CapturedPiece] shl 12 - PieceValues[TMove(movegen.Moves[j]).MovingPiece] >= xvalue;
      repeat inc(i) until PieceValues[TMove(movegen.Moves[i]).CapturedPiece] shl 12 - PieceValues[TMove(movegen.Moves[i]).MovingPiece] <= xvalue;
      if i<j then    // and (Compare(TMove(Moves[i]), TMove(Moves[j])) <> 0)
      begin
        if PieceValues[TMove(movegen.Moves[j]).CapturedPiece] shl 12 - PieceValues[TMove(movegen.Moves[j]).MovingPiece] <>
          PieceValues[TMove(movegen.Moves[i]).CapturedPiece] shl 12 - PieceValues[TMove(movegen.Moves[i]).MovingPiece]
        then
          movegen.SwapMoves(i,j)
      end else begin
        Result := j;
        exit;
      end;
    end;
  end;

  // this is the sorting itself
  procedure QuickSortCaptures(p, r: integer);
  var
    q: integer;
  begin
    if p < r then
    begin
      q := PartitionCaptures(p, r);
      QuickSortCaptures(p, q);
      QuickSortCaptures(q + 1, r);
    end;
  end;

  // almost the same function for sorting pieceful moves,
  // but looking on the factor of the hsitorical heuristics.
  function PartitionNormal(p, r: integer): integer;
  var
    i, j: integer;
    x: TMove;
    hvalue: integer;
  begin
    i := p - 1;
    j := r + 1;
    x := TMove(movegen.Moves[p]);
    hvalue := HistoryTable.History[CurrentPlayer][x.SourceSquare][x.DestinationSquare];
    while true do
    begin
      repeat dec(j) until HistoryTable.History[CurrentPlayer][TMove(movegen.Moves[j]).sourceSquare][TMove(movegen.Moves[j]).DestinationSquare] >= hvalue;
      repeat inc(i) until HistoryTable.History[CurrentPlayer][TMove(movegen.Moves[i]).sourceSquare][TMove(movegen.Moves[i]).DestinationSquare] <= hvalue;
      if i<j then
      begin
        // change the placs only in the case when the values are different
        if HistoryTable.History[CurrentPlayer][TMove(movegen.Moves[j]).sourceSquare][TMove(movegen.Moves[j]).DestinationSquare] <>
          HistoryTable.History[CurrentPlayer][TMove(movegen.Moves[i]).sourceSquare][TMove(movegen.Moves[i]).DestinationSquare]
        then
          movegen.SwapMoves(i,j)
      end else begin
        Result := j;
        exit;
      end;
    end;
  end;

  procedure QuickSortNormal(p, r: integer);
  var
    q: integer;
  begin
    if p < r then
    begin
      q := PartitionNormal(p, r);
      QuickSortNormal(p, q);
      QuickSortNormal(q + 1, r);
    end;
  end;

var
  bestIndex: integer;
  startCaptureIndex, endCaptureIndex, startNoncaptureIndex, endNoncaptureIndex,
  i, j: integer;

begin
  // get out bestMoveFountInTT  to the beginning of the list
  // with a great probability this move will be the best move here.
  startCaptureIndex := 0;
  if bestMoveFountInTT <> nil then
  begin
    bestIndex := movegen.FindIndex(bestMoveFountInTT);
    if bestIndex >= 0 then
    begin
      movegen.PullToFrontOfList(bestIndex);
      startCaptureIndex := 1;
    end else begin
      if (quiescence = false) and (bestMoveFountInTT.MoveType = MOVE_NORMAL) then
      begin
        raise Exception.Create('Cannot find among moves in non-quiescence mode !');
      end;
    end;
  end;

  // put all captures BEFORE the pieceful moves
  i := startCaptureIndex;

  if quiescence = false then
  begin
    // if this is not quiscense search, then there are pieceful moves
    // anf we need to consider them at last
    j := movegen.Moves.Count-1;
    while i<j do
    begin
      while (i<j) and (TMove(movegen.Moves[j]).MoveType = MOVE_NORMAL) do dec(j);
      while (i<j) and (TMove(movegen.Moves[i]).MoveType <> MOVE_NORMAL) do inc(i);
      movegen.SwapMoves(i, j);
    end;

    if TMove(movegen.Moves[j]).MoveType = MOVE_NORMAL then
      endCaptureIndex := j-1
    else
      endCaptureIndex := j;
  end
  else
    endCaptureIndex := movegen.Moves.Count - 1;

  startNonCaptureIndex := endCaptureIndex + 1;
  endNonCaptureIndex := movegen.Moves.Count - 1;

  // here we already know the location of captures and pieceful moves
  // in the arrays
  QuickSortCaptures(startCaptureIndex, endCaptureIndex); // sort captures by the type of captured piece
  QuickSortNormal(startNoncaptureIndex, endNonCaptureIndex); // and sort non-captures by the values of the history table
end;


// MTD(f)
function TAISearchAgent.MTDF( theBoard: TBoard; target, depth: integer; var BestMove: TMove): integer;
var
  beta: integer;
  currentEstimate, upperbound, lowerbound: integer;
begin
  currentEstimate := target;

  upperbound := ALPHABETA_MAXVAL; // f-
  lowerbound := ALPHABETA_MINVAL; // f+

  repeat
    beta := currentEstimate;
    if currentEstimate = lowerbound then inc(beta);

    // the call of alpha-beta with 0-window
    currentEstimate := AlphaBeta(MAXNODE, theBoard, depth, beta-1, beta, {var} bestMove, 0 );

    // correction of f-  f+
    if currentEstimate < beta then
      upperbound := currentEstimate
    else
      lowerbound := currentEstimate;
  until lowerbound >= upperbound;

  Result := currentEstimate;
end;


// ID-MTD(f)
function TAISearchAgent.PickBestMove(theBoard: TBoard): TMove;
var
  openingBookMove, bestMove: TMove;
  previousEstimate, iterdepth: integer;
  evalLower, evalUpper: integer;
begin
  bestMove := TMove.Create;

  // trying to find a move in the table of opening book.
  openingBookMove := OpeningBook.Query( theBoard );
  if  openingBookMove <> nil then
  begin
    bestMove.Clone(openingBookMove);
    Result := bestMove;
    exit;
  end;

  // the estimation is done from the viewpoing of CurrentPlayer
  FromWhosePerspective := theBoard.CurrentPlayer;

  previousEstimate := 0; // we can assign this, but this is stupid;
  // we have already probably searched starting from this node
  // therefore its useful to choose the middle of the bounds
  if TransTable.LookupBoardValuesOnly(theBoard, evalLower, evalUpper) then
    if (evalUpper < ALPHABETA_MAXVAL) or (evalLower > ALPHABETA_MINVAL) then
    begin
      if (evalUpper < ALPHABETA_MAXVAL) and (evalLower > ALPHABETA_MINVAL) then
        previousEstimate := (evalUpper + evalLower) div 2
      else if (evalUpper >= ALPHABETA_MAXVAL) then //we  have just lower value
        previousEstimate := evalLower + EVAL_THRESHOLD div 2
      else //we have just upper value
        previousEstimate := evalUpper - EVAL_THRESHOLD div 2;
    end;

  // clear redunrdant entris of TT
  TransTable.ClearRedundantEntries(theBoard);
  iterdepth := 0;

  while true do
  begin
    inc(iterdepth, 2); // increase the deepness

    // call MTD-f
    previousEstimate := MTDF( theBoard, previousEstimate, iterdepth, bestMove );

    write( 'Iteration of depth ', iterdepth, '; best move := ' );  bestMove.Print(); writeln( '  value = ', previousEstimate);
    if  iterdepth >= MAX_ITER_DEPTH then
      break;
  end;

  Result := bestMove;
end;



end.

