//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// IBChess
// Copyright (C) 2002 by Ihor Bobak, ibobak@ukrpost.net
//------------------------------------------------------------------------------
// evaluation function
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
unit boardeval;

interface

uses global, board;

type
  TBoardEvaluator = class
  private
    // quantities of pawns on the files
    PawnCountOnFiles: array[SIDE_WHITE..SIDE_BLACK, 0..7] of integer;
    // quantities of pawns on the white and black squares
    PawnCountOnSruareColor: array[SIDE_WHITE..SIDE_BLACK, 0..1] of integer;
    // total number of ppawns for each side. 
    TotalPawns: array[SIDE_WHITE..SIDE_BLACK] of integer;
    // the values (for each color and each file) of maximum and minimum
    // rank where the pawn is located
    MostAdvancedRank, LeastAdvancedRank: array[SIDE_WHITE..SIDE_BLACK, 0..7] of integer;
    // pawn structure anallysis; used in EvalPawnStructure
    procedure AnalyzePawnStructure(theBoard: TBoard);
    // the estimation of the pawn position
    function EvalPawnStructure(side: integer): integer;
    function EvalPawnStructureDiff(FromwhosePerspective: integer): integer;
    // the estimation of each piece location
    function EvalWhitePieceDevelopment(theBoard: TBoard): integer;
    function EvalBlackPieceDevelopment(theBoard: TBoard): integer;
    function EvalPieceDevelopmentDiff(theBoard: TBoard; side: integer): integer;
    // bonuses for the open rooks
    function EvalRookBonus(theBoard: TBoard; side: integer): integer;
    function EvalRookBonusDiff(theBoard: TBoard; FromWhosePerspective: integer): integer;
    // the castling estimation
    function EvalCastling(theBoard: TBoard; side: integer): integer;
    function EvalCastlingDiff(theBoard: TBoard; fromWhosePerspective: integer): integer;
    // penalties for the bad bishop position
    function EvalBadBishops(theBoard: TBoard; side: integer): integer;
    function EvalBadBishopsDiff(theBoard: TBoard; fromWhosePerspective: integer): integer;
    // the estimation of the king defence
    function EvalKingSafety(theBoard: TBoard; side: integer): integer;
    function EvalKingSafetyDiff(theBoard: TBoard; fromWhosePerspective: integer): integer;
  public
    // the function which returns the final estimation
    function Evaluate(theBoard: TBoard; fromWhosePerspective: integer): integer;
  end;

implementation

const
  // the array of benefits from the location of each kind
  // of piece on each square
  PositionalAdvantage: array[WHITE_PAWN..BLACK_KING] of array[0..63] of integer =
  (
    // white pawn
    (0,  0,  0,  0,  0,  0,  0,  0,
    10, 15, 24, 32, 32, 24, 15, 10,
    10, 15, 24, 32, 32, 24, 15, 10,
    6,  11, 18, 27, 27, 16, 11, 6,
    4,  7,  18, 25, 25, 16, 7,  4,
    4,  6,  12, 15, 15, 4,  6,  4,
    2,  3,  4,  8,  8,  4,  3,  2,
    0,  0,  0,  0,  0,  0,  0,  0),
    // black pawn
    (0,  0,  0,  0,  0,  0,  0,  0,
    2,  3,  4,  8,  8,  4,  3,  2,
    4,  6,  12, 15, 15, 4,  6,  4,
    4,  7,  18, 25, 25, 16, 7,  4,
    6,  11, 18, 27, 27, 16, 11, 6,
    10, 15, 24, 32, 32, 24, 15, 10,
    10, 15, 24, 32, 32, 24, 15, 10,
    0,  0,  0,  0,  0,  0,  0,  0),
    // white knight
    (-7, -3, 1,  3,  3,  1,  -3, -7,
    2,  6,  14, 20, 20, 14, 6,  2,
    6,  14, 28, 32, 32, 28, 14, 6,
    8,  18, 30, 32, 32, 30, 18, 8,
    8,  18, 26, 30, 30, 26, 18, 8,
    6,  14, 22, 26, 26, 22, 14, 6,
    2,  6,  14, 20, 20, 14, 6,  2,
    -7, -3, 1,  3,  3,  1,  -3, -7),
    // black knight
    (-7, -3, 1,  3,  3,  1,  -3, -7,
    2,  6,  14, 20, 20, 14, 6,  2,
    6,  14, 22, 26, 26, 22, 14, 6,
    8,  18, 26, 30, 30, 26, 18, 8,
    8,  18, 30, 32, 32, 30, 18, 8,
    6,  14, 28, 32, 32, 28, 14, 6,
    2,  6,  14, 20, 20, 14, 6,  2,
    -7, -3, 1,  3,  3,  1,  -3, -7),
    // white bishop
    (16, 16, 16, 16, 16, 16, 16, 16,
    16, 29, 31, 31, 31, 31, 29, 16,
    16, 28, 32, 32, 32, 32, 28, 16,
    16, 26, 32, 32, 32, 32, 26, 16,
    16, 26, 32, 32, 32, 32, 26, 16,
    26, 28, 32, 32, 32, 32, 28, 26,
    26, 29, 31, 31, 31, 31, 29, 26,
    16, 16, 16, 16, 16, 16, 16, 16),
    // black bishop
    (16, 16, 16, 16, 16, 16, 16, 16,
    26, 29, 31, 31, 31, 31, 29, 26,
    26, 28, 32, 32, 32, 32, 28, 26,
    16, 26, 32, 32, 32, 32, 26, 16,
    16, 26, 32, 32, 32, 32, 26, 16,
    16, 28, 32, 32, 32, 32, 28, 16,
    16, 29, 31, 31, 31, 31, 29, 16,
    16, 16, 16, 16, 16, 16, 16, 16),
    // white rook
    (0,  0,  0,  0,  0,  0,  0,  0,
    10, 10, 10, 10, 10, 10, 10, 10,
    -2, 0,  0,  0,  0,  0,  0,  -2,
    -2, 0,  0,  0,  0,  0,  0,  -2,
    -2, 0,  0,  0,  0,  0,  0,  -2,
    -2, 0,  0,  0,  0,  0,  0,  -2,
    -2, 0,  0,  0,  0,  0,  0,  -2,
    0,  0,  0,  3,  3,  0,  0,  0),
    // black rook
    (0,  0,  0,  3,  3,  0,  0,  0,
    -2, 0,  0,  0,  0,  0,  0,  -2,
    -2, 0,  0,  0,  0,  0,  0,  -2,
    -2, 0,  0,  0,  0,  0,  0,  -2,
    -2, 0,  0,  0,  0,  0,  0,  -2,
    -2, 0,  0,  0,  0,  0,  0,  -2,
    10, 10, 10, 10, 10, 10, 10, 10,
    0,  0,  0,  0,  0,  0,  0,  0),
    // white queen
    (-2, -2, 0,  0,  0,  0,  0,  0,
    -2, -2, 0,  0,  0,  0,  0,  0,
    -2, -2, 0,  0,  0,  0,  0,  0,
    0,  0,  0,  2,  2,  0,  0,  0,
    0,  0,  0,  2,  2,  0,  0,  0,
    0,  1,  1,  1,  1,  0,  0,  0,
    0,  0,  1,  1,  1,  0,  0,  0,
    -2, -2, -2, 0,  0,  -2, -2, -2),
    // black queen
    (-2, -2, -2, 0,  0,  -2, -2, -2,
    0,  0,  1,  1,  1,  0,  0,  0,
    0,  1,  1,  1,  1,  0,  0,  0,
    0,  0,  0,  2,  2,  0,  0,  0,
    0,  0,  0,  2,  2,  0,  0,  0,
    -2, -2, 0,  0,  0,  0,  0,  0,
    -2, -2, 0,  0,  0,  0,  0,  0,
    -2, -2, 0,  0,  0,  0,  0,  0),
    // white king
    (-20,-20,-20,-20,-20,-20,-20,-20,
    -20,-20,-20,-20,-20,-20,-20,-20,
    -20,-20,-20,-20,-20,-20,-20,-20,
    -20,-20,-20,-20,-20,-20,-20,-20,
    -20,-20,-20,-20,-20,-20,-20,-20,
    -7, -7, -7, -9, -9, -7, -7, -7,
    -5, -5, -5, -5, -5, -5, -5, -5,
    1,  3,  5,  -8, -5, -8, 5,  2),
    // black king
    (1,  3,  5,  -8, -5, -8, 5,  2,
    -5, -5, -5, -5, -5, -5, -5, -5,
    -7, -7, -7, -9, -9, -7, -7, -7,
    -20,-20,-20,-20,-20,-20,-20,-20,
    -20,-20,-20,-20,-20,-20,-20,-20,
    -20,-20,-20,-20,-20,-20,-20,-20,
    -20,-20,-20,-20,-20,-20,-20,-20,
    -20,-20,-20,-20,-20,-20,-20,-20)
  );


function TBoardEvaluator.Evaluate(theBoard: TBoard; fromWhosePerspective: integer): integer;
var
  evalMat, evalPos: integer;
begin
  // the evaluation of the material part is done in the board class
  evalMat := theBoard.EvalMaterialDiff(fromWhosePerspective);

  // pawn analysis
  AnalyzePawnStructure(theBoard);
  // the sum of all material factors
  evalPos :=
       EvalPawnStructureDiff(fromWhosePerspective) +
       EvalPieceDevelopmentDiff(theBoard, fromWhosePerspective)+
       EvalRookBonusDiff(theBoard, FromWhosePerspective) +
       EvalCastlingDiff(theBoard, fromWhosePerspective) +
       EvalBadBishopsDiff(theBoard, fromWhosePerspective) +
       EvalKingSafetyDiff(theBoard, fromWhosePerspective);
  // the summary evaluation
  Result := evalMat + evalPos;
end;

// pawn structure analysis
procedure TBoardEvaluator.AnalyzePawnStructure(theBoard: TBoard);
var
  i, square, rank, file1: integer;
begin
  fillchar(PawnCountOnFiles, sizeof(PawnCountOnFiles), 0);
  fillchar(PawnCountOnSruareColor, sizeof(PawnCountOnSruareColor), 0);
  fillchar(TotalPawns, sizeof(TotalPawns), 0);
  for i := 0 to 7 do
  begin
    MostAdvancedRank[SIDE_WHITE, i] := 7; //minimum value
    LeastAdvancedRank[SIDE_BLACK, i] := 7;
    MostAdvancedRank[SIDE_BLACK, i] := 0; //maximum value
    LeastAdvancedRank[SIDE_WHITE, i] := 0;
  end;

  // white pawns analysis
  for square := 55 downto 8 do
  begin
    rank := square shr 3;
    file1 := square mod 8;
    if theBoard.BitBoards[WHITE_PAWN] and SquareBits[square] <> 0 then
    begin
      inc(PawnCountOnFiles[SIDE_WHITE, file1]);
      inc(TotalPawns[SIDE_WHITE]);
      // find the pawn which is maximally promoted
      MostAdvancedRank[SIDE_WHITE, file1] := rank;
      if (rank mod 2) = (file1 mod 2) then
        inc(PawnCountOnSruareColor[SIDE_WHITE, 0])
      else
        inc(PawnCountOnSruareColor[SIDE_WHITE, 1]);
    end
    else
    // find the minimally promoted pawn
    if theBoard.BitBoards[BLACK_PAWN] and SquareBits[square] <> 0 then
      LeastAdvancedRank[SIDE_BLACK, file1] := rank;
    // pay attention at the order of the pawn consideration
  end;

  // the analysis of the black pawns - by analogy
  for square := 8 to 55 do
  begin
    rank := square shr 3;
    file1 := square mod 8;
    if theBoard.BitBoards[BLACK_PAWN] and SquareBits[square] <> 0 then
    begin
      inc(PawnCountOnFiles[SIDE_BLACK, file1]);
      inc(TotalPawns[SIDE_BLACK]);
      MostAdvancedRank[SIDE_BLACK, file1] := rank;
      if (rank mod 2) = (file1 mod 2) then
        inc(PawnCountOnSruareColor[SIDE_BLACK, 0])
      else inc(PawnCountOnSruareColor[SIDE_BLACK, 1]);
    end
    else if theBoard.BitBoards[WHITE_PAWN] and SquareBits[square] <> 0 then
      LeastAdvancedRank[SIDE_WHITE, file1] := rank;
  end;
end;

// the estimation of the pawns from the player "side" perspective
function TBoardEvaluator.EvalPawnStructure(side: integer): integer;
var
  otherside, score, i: integer;
begin
  score := 0;
  otherside := (side + 1) mod 2;

  // penalties for coupled pawns
  for i := 0 to 7 do
    if PawnCountOnFiles[side, i] > 1 then
      dec(score, PENALTY_DOUBLED_PAWN);

  // penalties for isolated pawns
  if (PawnCountOnFiles[side, 0] > 0) and (PawnCountOnFiles[side, 1] = 0) then dec(score, PENALTY_ISOLATED_PAWN);
  if (PawnCountOnFiles[side, 7] > 0) and (PawnCountOnFiles[side, 6] = 0) then dec(score, PENALTY_ISOLATED_PAWN);
  for i := 1 to 6 do
    if (PawnCountOnFiles[side, i] > 0) and (PawnCountOnFiles[side, i - 1] = 0) and (PawnCountOnFiles[side, i + 1] = 0) then
      dec(score, PENALTY_ISOLATED_PAWN);

  // small penalty for the presence of all pawns on the board
  if TotalPawns[side] = 8 then
    dec(score, PENALTY_ALL_PAWNS_ALIVE);

  // bonuses for the pawns which are able to pass to the last rank
  if side = SIDE_WHITE then
  begin
    if MostAdvancedRank[side, 0] <= min(LeastAdvancedRank[otherside, 0], LeastAdvancedRank[otherside, 1]) then
      inc(score, sqr(6 - MostAdvancedRank[side, 0]));
    if MostAdvancedRank[side, 7] <= min(LeastAdvancedRank[otherside, 7], LeastAdvancedRank[otherside, 6]) then
      inc(score, sqr(6 - MostAdvancedRank[side, 7]));
    for i := 1 to 6 do
      if (MostAdvancedRank[side, i] <= LeastAdvancedRank[otherside, i]) and
         (MostAdvancedRank[side, i] <= LeastAdvancedRank[otherside, i - 1]) and
         (MostAdvancedRank[side, i] <= LeastAdvancedRank[otherside, i + 1]) then
        inc(score, sqr(6 - MostAdvancedRank[side, i]));
  end
  else // for black pawns
  begin
    if MostAdvancedRank[side, 0] >= max(LeastAdvancedRank[otherside, 0], LeastAdvancedRank[otherside, 1]) then
      inc(score, sqr(MostAdvancedRank[side, 0]-1));
    if MostAdvancedRank[side, 7] >= max(LeastAdvancedRank[otherside, 7], LeastAdvancedRank[otherside, 6]) then
      inc(score, sqr(MostAdvancedRank[side, 7]-1));
    for i := 1 to 6 do
      if (MostAdvancedRank[side, i] >= LeastAdvancedRank[otherside, i]) and
         (MostAdvancedRank[side, i] >= LeastAdvancedRank[otherside, i - 1]) and
         (MostAdvancedRank[side, i] >= LeastAdvancedRank[otherside, i + 1]) then
        inc(score, sqr(MostAdvancedRank[side, i]-1));
  end;

  Result := score;
end;


// the difference is a component of the evaluation function
function TBoardEvaluator.EvalPawnStructureDiff(FromwhosePerspective: integer): integer;
var
  opponent: integer;
begin
  opponent := (FromWhosePerspective + 1) mod 2;
  Result := EvalPawnStructure(FromWhosePerspective) - EvalPawnStructure(opponent);
end;

// the estimation of the advantage of the piece location (for white)
function TBoardEvaluator.EvalWhitePieceDevelopment(theBoard: TBoard): integer;
var
  i, score: integer;
begin
  score := 0;
  for i:=0 to 63 do
    if theBoard.BitBoards[ALL_WHITE_PIECES] and SquareBits[i] <> 0 then
      inc(score, PositionalAdvantage[theBoard.FindWhitePiece(i), i]);
  Result := score;
end;

// the estimation of the advantage of the piece location (for black)
function TBoardEvaluator.EvalBlackPieceDevelopment(theBoard: TBoard): integer;
var
  i, score: integer;
begin
  score := 0;
  for i:=0 to 63 do
    if theBoard.BitBoards[ALL_BLACK_PIECES] and SquareBits[i] <> 0 then
      inc(score, PositionalAdvantage[theBoard.FindBlackPiece(i), i]);
  Result := score;
end;

// the difference between the previous 2 values
function TBoardEvaluator.EvalPieceDevelopmentDiff(theBoard: TBoard; side: integer): integer;
var
  score: integer;
begin
  score := EvalWhitePieceDevelopment(theBoard) -  EvalBlackPieceDevelopment(theBoard);
  if side = SIDE_BLACK then
    score := - score;
  Result := score;
end;

// the king defence; this is not done till the end, since I need
// to perform more investigation in this area.
function TBoardEvaluator.EvalKingSafety(theBoard: TBoard; side: integer): integer;
var
  score: integer;
begin
  score := 0;
  // will be developed in the future
  Result := score;
end;


function TBoardEvaluator.EvalKingSafetyDiff(theBoard: TBoard; fromWhosePerspective: integer): integer;
var
  opponent: integer;
begin
  opponent := (fromWhosePerspective + 1) mod 2;
  Result := EvalKingSafety(theBoard, fromWhosePerspective) - EvalKingSafety(theBoard, opponent);
end;

// the rook position bonuses
function TBoardEvaluator.EvalRookBonus(theBoard: TBoard; side: integer): integer;
var
  rookboard: TBitBoard;
  score, square, file1, otherside: integer;
begin
  score := 0;
  otherside := (side + 1) mod 2;

  rookboard := theBoard.BitBoards[ROOK+side];
  if rookboard = 0 then
  begin
    Result := 0;
    exit;
  end;

  for square := 0 to 63 do
    if rookboard and SquareBits[square] <> 0 then
    begin
      file1 := square mod 8 ;
      // if there are no our pawns on the file,
      // its easier to pass there for the rook
      if PawnCountOnFiles[side, file1] = 0 then
        // if there are no opponent pawns also,
        // we give more bonuses
        if PawnCountOnFiles[otherside, file1] = 0 then
          inc(score, ENCOURAGE_ROOK_OPEN)
        else inc(score, ENCOURAGE_ROOK_SEMIOPEN);
    end;
  Result := score;
end;


function TBoardEvaluator.EvalRookBonusDiff(theBoard: TBoard; FromWhosePerspective: integer): integer;
var
  opponent: integer;
begin
  opponent := (FromWhosePerspective + 1) mod 2;
  Result := EvalRookBonus(theBoard, FromWhosePerspective) - EvalRookBonus(theBoard, opponent);
end;


// bonuses/penalties for rocking
function TBoardEvaluator.EvalCastling(theBoard: TBoard; side: integer): integer;
var
  score, otherside: integer;
begin
  score := 0;
  otherside := (side + 1) mod 2;

  // if the opponents queen is on the board, then
  if theBoard.BitBoards[QUEEN+otherside] <> 0 then
  begin
    // if the rocking is done - bonuses,
    if  theBoard.HasCastled[side] then
      inc(score, ENCOURAGE_HAS_CASTLED)
    // if not already done, then if there is both sides possibilites
    else if theBoard.CastlingPossibility[side + CASTLE_QUEENSIDE] and
            theBoard.CastlingPossibility[side + CASTLE_KINGSIDE] then
      dec(score, PENALTY_HADNT_CASTLED_11) // then a little penaltyy
    // if it is not done, then check the kings's side
    else if theBoard.CastlingPossibility[side + CASTLE_KINGSIDE] then
      dec(score, PENALTY_HADNT_CASTLED_01)
    else if theBoard.CastlingPossibility[side + CASTLE_QUEENSIDE] then
      dec(score, PENALTY_HADNT_CASTLED_10)
    else
      dec(score, PENALTY_HADNT_CASTLED_00);
  end;
  Result := score;
end;


function TBoardEvaluator.EvalCastlingDiff(theBoard: TBoard; fromWhosePerspective: integer): integer;
var
  opponent: integer;
begin
  opponent := (FromWhosePerspective + 1) mod 2;
  Result := EvalCastling(theBoard, fromWhosePerspective) - EvalCastling(theBoard, opponent);
end;

// penalties for badly located bishops
function TBoardEvaluator.EvalBadBishops(theBoard: TBoard; side: integer): integer;
var
  bishopsBoard: TBitBoard;
  score, square, rank, file1: integer;
begin
  bishopsBoard := theBoard.BitBoards[BISHOP + side];
  if bishopsBoard = 0 then
  begin
    Result := 0;
    exit;
  end;

  // take a bishop, count the pawns on the same square color,
  // and this quantity will be the penalty measure
  score := 0;
  for square := 0 to 63 do
  begin
    if bishopsBoard and SquareBits[square] <> 0 then
    begin
      rank := square shr 3 ;
      file1 := square mod 8;
      if ((rank mod 2) = (file1 mod 2)) then
        dec(score,  PawnCountOnSruareColor[side, 0])
      else
        dec(score,  PawnCountOnSruareColor[side, 1]);

      bishopsBoard := bishopsBoard xor SquareBits[square];
      if bishopsBoard = 0 then
        break;
    end;
  end;
  Result := score;
end;


function TBoardEvaluator.EvalBadBishopsDiff(theBoard: TBoard; fromWhosePerspective: integer): integer;
var
  opponent: integer;
begin
  opponent := (fromWhosePerspective + 1) mod 2;
  Result := EvalBadBishops(theBoard, fromWhosePerspective) - EvalBadBishops(theBoard, opponent);
end;

end.
