//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// IBChess
// Copyright (C) 2002 by Ihor Bobak, ibobak@ukrpost.net
//------------------------------------------------------------------------------
// transposition table and history heuristics
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
unit tables;

interface

uses global, board,  move,  movegenerator;

const
  TABLE_SIZE = 1024*1024 div 2;

type
  // the cell of the transposition tabl
  TTranspositionEntry = class
  public
    EmptyFlag: boolean; // true = empty
    theLock: TBitBoard; // the hash-lock value for collision prevention
    EvaluationLower: integer; // the lower bound for position estimations
    EvaluationUpper: integer; // the upper bound for position estimations
    SearchedDepth: integer; // the deepness to which this position was developer
    theBestMove: TMove; // the best move from this postions
    TotalPiecesByColor: array[0..1] of integer; // how many pieces are for each color
    constructor Create;
    destructor Destroy; override;
  end;

type
  // the TT
  TTranspositionTable = class
  private
    Table: array of TTranspositionEntry; // TT
  public
    constructor Create;
    destructor Destroy; override;
    // search of the position in the table
    function LookupBoard(theBoard: TBoard; var aEvaluationLower, aEvaluationUpper: integer; var lookupMov: TMove; var aSearchedDepth: integer): boolean;
    // the helper function using the previous one
    function LookupBoardValuesOnly(theBoard: TBoard; var aEvaluationLower, aEvaluationUpper: integer): boolean;
    // the storage of the position into the table
    function StoreBoard(theBoard: TBoard; aEvaluationLower, aEvaluationUpper, aDepth: integer; aBestMove: TMove): boolean;
    // cleaning of unncecessary positions of the board
    procedure ClearRedundantEntries(currentBoard: TBoard);
  end;


type
  // gistorical heuristics
  THistoryTable = class
  public
    History: array[SIDE_WHITE..SIDE_BLACK, 0..ALL_SQUARES-1, 0..ALL_SQUARES-1] of integer;
    constructor Create;
    // increments the corresponding item
    function AddCount( whichPlayer: integer; mov: TMove; depth: integer): boolean;
  end;

implementation

//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// TTranspositionEntry
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
constructor TTranspositionEntry.Create;
begin
  EmptyFlag := true;
  EvaluationLower := ALPHABETA_MINVAL;
  EvaluationUpper := ALPHABETA_MAXVAL;
  theBestMove := TMove.Create;
end;

destructor TTranspositionEntry.Destroy;
begin
  theBestMove.Free;
end;

//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// TTranspositionTable
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
constructor TTranspositionTable.Create;
var
  i: integer;
begin
  SetLength(Table, TABLE_SIZE);
  for i := 0 to TABLE_SIZE -1 do
    Table[i] := TTranspositionEntry.Create;
end;

destructor TTranspositionTable.Destroy;
var
  i: integer;
begin
  for i := 0 to TABLE_SIZE -1 do
    Table[i].Free;
end;


// search of the position in the table
function TTranspositionTable.LookupBoard(
  theBoard: TBoard;
  var aEvaluationLower, aEvaluationUpper: integer;
  var lookupMov: TMove;
  var aSearchedDepth: integer): boolean;
var
  key: integer;
  entry: TTranspositionEntry;
begin
  Result := false;

  key := abs(theBoard.HashKey mod TABLE_SIZE); // find the key
  entry := Table[key];

  if entry.EmptyFlag then // if the entry is empty, then exit
    exit;

  if entry.theLock <> theBoard.HashLock then // in the case of the collision
  // (when the key is the same but the lock is not), go out
    exit;

  // return the found result
  lookupMov.Clone(entry.theBestMove);
  aEvaluationLower := entry.EvaluationLower;
  aEvaluationUpper := entry.EvaluationUpper;
  aSearchedDepth   := entry.SearchedDepth;

  Result := true;
end;

// does the same as the previous function
// but does not return the best move
function TTranspositionTable.LookupBoardValuesOnly(theBoard: TBoard; var aEvaluationLower, aEvaluationUpper: integer): boolean;
var
  lookupMove: TMove;
  aSearchedDepth: integer;
begin
  lookupMove := TMove.Create;
  Result := LookupBoard(theBoard, aEvaluationLower, aEvaluationUpper,  lookupMove, aSearchedDepth);
  lookupMove.Free;
end;

// writes down the position to the table
function TTranspositionTable.StoreBoard(
  theBoard: TBoard;
  aEvaluationLower, aEvaluationUpper,
  aDepth: integer;
  aBestMove: TMove): boolean;
var
  key, lock: integer;
  entry: TTranspositionEntry;
begin
  Result := false;
  // counting the key and the lock
  key  := abs( theBoard.HashKey mod TABLE_SIZE );
  lock := theBoard.HashLock;
  entry := Table[key];

  if entry.EmptyFlag or // if the entry is empty
    // or occupied by the same position (equal lock),
    // but the seach deepness was less,
    // then hold the new values of the search window
    ((lock = entry.theLock) and (aDepth > entry.SearchedDepth)) then
  begin
    with entry do
    begin
      EmptyFlag := false;
      theLock := lock;
      theBestMove.Clone(aBestMove);
      EvaluationLower := aEvaluationLower;
      EvaluationUpper := aEvaluationUpper;
      SearchedDepth   := aDepth;
      TotalPiecesByColor[0] := theBoard.TotalPiecesByColor[0];
      TotalPiecesByColor[1] := theBoard.TotalPiecesByColor[1];
    end;
    Result := true;
  end;
end;


// deletes the redundant records from TT
// if at some time there are m white and n black pieces, then
// in the future there CANNOT be positions where there are more
// then m or n pieces of the corresponding color.
// We don't need this memory in the TT
procedure TTranspositionTable.ClearRedundantEntries(currentBoard: TBoard);
var
  i, whitePieceCount, blackPieceCount: integer;
begin
  whitePieceCount := currentBoard.TotalPiecesByColor[0];
  blackPieceCount := currentBoard.TotalPiecesByColor[1];
  for i:=0 to TABLE_SIZE-1 do
    if (Table[i].TotalPiecesByColor[0] > whitePieceCount) or
       (Table[i].TotalPiecesByColor[1] > blackPieceCount) then
     Table[i].EmptyFlag := true;
end;


//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// THistoryTable
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
constructor THistoryTable.Create;
begin
  fillchar(history, sizeof(History), 0);
end;

// the increment of moves counter
function THistoryTable.AddCount( whichPlayer: integer; mov: TMove; depth: integer): boolean;
begin
  inc(History[whichPlayer][mov.SourceSquare][mov.DestinationSquare], 1 shl depth);
  Result := true;
end;

end.

